/*
 *  Copyright (C) 2004-2008  gulikoza
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

// DXplay.cpp : implementation file
//

#include "DXplay.h"

#include <atlbase.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CDXplay
#ifdef _DEBUG
#define SAFE_RELEASE(p, s) { if(p) { wsprintf(debug, TEXT("DXplay: Releasing %s, refcount: %d"), s, p->Release()); OutputDebugString(debug); p = NULL; } \
								else { wsprintf(debug, TEXT("DXplay: ERROR:Trying to release a NULL pointer %s"), s); OutputDebugString(debug); } }
#else
#define SAFE_RELEASE(p, s) { if(p) { p->Release(); p = NULL; } }
#endif

static const CLSID CLSID_Mpeg2Dec = {0x39F498AF, 0x1A09, 0x4275, {0xB1, 0x93, 0x67, 0x3B, 0x0B, 0xA3, 0xD4, 0x78}};

// Debug functions to view the graph remotely
#ifdef _DEBUG
void RemoveFromRot(DWORD pdwRegister)
{
    IRunningObjectTable *pROT;
    if(SUCCEEDED(GetRunningObjectTable(0, &pROT))) {
        pROT->Revoke(pdwRegister);
        pROT->Release();
    }
}

HRESULT AddToRot(IUnknown *pUnkGraph, DWORD *pdwRegister) 
{
    IMoniker * pMoniker;
    IRunningObjectTable *pROT;
    if(FAILED(GetRunningObjectTable(0, &pROT))) {
        return E_FAIL;
    }
    WCHAR wsz[256];
    wsprintfW(wsz, L"FilterGraph %08x pid %08x", (DWORD)pUnkGraph, GetCurrentProcessId());

    HRESULT hr = CreateItemMoniker(L"!", wsz, &pMoniker);
    if(SUCCEEDED(hr)) {
        hr = pROT->Register(ROTFLAGS_REGISTRATIONKEEPSALIVE, pUnkGraph,
            pMoniker, pdwRegister);
        pMoniker->Release();
    }
    pROT->Release();
    return hr;
}
#endif


CDXplay::CDXplay()
{
	//Initialize

	pGB = NULL;
	pMC = NULL;
	pVW = NULL;
	pME = NULL;
	pBA = NULL;

	hEvent = 0;
}


//This creates the filter graph manager
void CDXplay::BuildFilterGraph(void)
{
	if(CoCreateInstance(CLSID_FilterGraph, NULL, CLSCTX_INPROC,IID_IGraphBuilder, (void **)&pGB) != S_OK)
	{
		OUTPUTDEBUG("DXplay: Graph NOT built!");
		pGB = NULL;
		return;
	}

#ifdef _DEBUG
	AddToRot(pGB, &dwRegister);
#endif

	//Query COM interface
	pGB->QueryInterface(IID_IMediaControl, (void **)&pMC);
	pGB->QueryInterface(IID_IVideoWindow, (void **)&pVW);
	pGB->QueryInterface(IID_IMediaEventEx, (void **)&pME);
	pGB->QueryInterface(IID_IBasicAudio, (void**)&pBA);

#ifdef _DEBUG
	char	logfilename[MAX_PATH];

	// create win32 file handle
	wsprintf(logfilename, TEXT("c:\\graph.log"));
	logfile = CreateFile(logfilename, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);

	pGB->SetLogFile((unsigned long)logfile);
#endif

	OUTPUTDEBUG("DXplay: BuildFilterGraph completed.");
}

CDXplay::~CDXplay()
{
	OUTPUTDEBUG("DXplay: Destructor called.");

	Delete();
}


BOOL CDXplay::RenderMediaFile(const char* pszFileName)
{
	IBaseFilter *pVmr, *g_pMpeg2Dec;

#ifdef _DEBUG
	wsprintf(debug, TEXT("DXplay: RenderMediaFile called: %s."), pszFileName);
	OutputDebugString(debug);
#endif

	HRESULT hr;

	USES_CONVERSION;

	if(!pGB)
		BuildFilterGraph();

	// Use VMR9 when requested
	if(bVMR9) {
		if(CoCreateInstance(CLSID_VideoMixingRenderer9, 0, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void**)&pVmr) != S_OK) {
			OUTPUTDEBUG("DXplay: Failed to create VMR9");
		} else {
			pGB->AddFilter(pVmr, L"VMR9");
		}
	}

	// Use Mpeg2Dec when requested
	if(bMpeg2Dec) {
		if(CoCreateInstance(CLSID_Mpeg2Dec, NULL, CLSCTX_INPROC_SERVER, IID_IBaseFilter, (void**)&g_pMpeg2Dec) != S_OK) {
			OUTPUTDEBUG("DXplay: Failed to create Mpeg2Dec filter");
		} else {
			pGB->AddFilter(g_pMpeg2Dec, L"Mpeg2Dec Filter");
		}
	}

	if(pGB)	{
		if(bAltPath) {
			char *c;
			c = strrchr(pszFileName, '\\');
			strcat(szPath, c);
#ifdef _DEBUG
			wsprintf(debug, TEXT("DXplay: Using alternate file path: %s."), szPath);
			OutputDebugString(debug);
#endif
			if((hr=pGB->RenderFile(A2W(szPath), NULL))==S_OK) {
				// Alternate path successful
				c = strrchr(szPath, '\\');
				*c = '\0';
				OUTPUTDEBUG("DXplay: RenderMediaFile completed.");
				return TRUE;
			}
			c = strrchr(szPath, '\\');
			*c = '\0';
			OUTPUTDEBUG("DXplay: Alternate file not found. Fallback to normal path...");
		}

		if((hr=pGB->RenderFile(A2W(pszFileName), NULL)) != S_OK) {
			ErrorReport(hr);
			OUTPUTDEBUG("DXplay: Error rendering file!");
			return FALSE;
		}
	} else {
		OUTPUTDEBUG("DXplay: ERROR! No graph!");
		return FALSE;
	}

	OUTPUTDEBUG("DXplay: RenderMediaFile completed.");

	return TRUE;
}


BOOL CDXplay::SetParentWindow(HWND m_hwnd)
{

#ifdef _DEBUG
	wsprintf(debug, TEXT("DXplay: SetParentWindow called: %d."), m_hwnd);
	OutputDebugString(debug);
#endif

	if(!pVW) {
		OUTPUTDEBUG("DXplay: ERROR! No window to attach!?");
		return FALSE;
	}
	
	SendMessage(m_hwnd, WM_ERASEBKGND, NULL, NULL);

	//Set the parents window
	pVW->put_Owner((OAHWND)m_hwnd);

	//Sets the child window
	pVW->put_WindowStyle(WS_CHILD | WS_CLIPSIBLINGS);

	//here we create a RECT in which to draw the child window on
	RECT vwrect;

	//get the size of the Parent
	GetClientRect(m_hwnd, &vwrect);

	//Set the Child to this position
	pVW->SetWindowPosition(0, 0, vwrect.right, vwrect.bottom);
	//pVW->SetWindowForeground(true);

	if(pME)	{

#ifdef _DEBUG
		wsprintf(debug, TEXT("DXplay: SetNotify window: %d."), m_hwnd);
		OutputDebugString(debug);
#endif

		pME->SetNotifyWindow((OAHWND)m_hwnd, WM_P_GRAPHNOTIFY, 0);
	}

	OUTPUTDEBUG("DXplay: SetParentWindow completed.");

	return TRUE;
}


BOOL CDXplay::Run()
{
	OUTPUTDEBUG("DXplay: Run called.");

	if(pMC)
		pMC->Run();
	else {
		OUTPUTDEBUG("DXplay: ERROR! No media control.");
		return FALSE;
	}

	OUTPUTDEBUG("DXplay: Run completed.");

	return TRUE;
}


BOOL CDXplay::Stop()
{
	OUTPUTDEBUG("DXplay: Stop called.");

	if(pMC)
		pMC->Stop();

	if(pVW)	{
		pVW->put_Visible(OAFALSE);
		pVW->put_Owner(NULL);
	}

	if (pME)
		pME->SetNotifyWindow(NULL, WM_P_GRAPHNOTIFY, NULL);

	if(tier >= T3_REMOVEALL)
		ReleaseAllFilters();

	OUTPUTDEBUG("DXplay: Stop completed.");

	return TRUE;
}


BOOL CDXplay::Delete()
{
	OUTPUTDEBUG("DXplay: Delete called.");

	Stop();
	SAFE_RELEASE(pMC, "pMC")
	SAFE_RELEASE(pVW, "pVW")
	SAFE_RELEASE(pME, "pME")
	SAFE_RELEASE(pBA, "pBA")

	if(pGB)	{
#ifdef _DEBUG
		pGB->SetLogFile(NULL);
		if(logfile)
			CloseHandle(logfile);
		RemoveFromRot(dwRegister);
#endif
		SAFE_RELEASE(pGB, "pGB")
	}

	OUTPUTDEBUG("DXplay: Delete completed.");

	return TRUE;
}


BOOL CDXplay::Pause()
{
	OUTPUTDEBUG("DXplay: Pause called.");

	if(pMC)
		pMC->Pause();
	else {
		OUTPUTDEBUG("DXplay: ERROR! No media control.");
		return FALSE;
	}

	OUTPUTDEBUG("DXplay: Pause completed.");

	return TRUE;
}


BOOL CDXplay::OnGraphEvent(void)
{
	long EventCode, Param1, Param2;

	OUTPUTDEBUG("DXplay: OnGraphEvent called.");

	if(!pME) {
		OUTPUTDEBUG("DXplay: ERROR! No media event!");
		return FALSE;
	}

	while(pME->GetEvent(&EventCode, &Param1, &Param2, 0) != E_ABORT) {
		pME->FreeEventParams(EventCode, Param1, Param2);

		if(EC_COMPLETE == EventCode) {
			pME->FreeEventParams(EventCode, Param1, Param2);
			OUTPUTDEBUG("DXplay: OnGraphEvent: Video complete.");
			return TRUE;			// Video is complete
		}
	}

	return FALSE;
}


// Function returns FileSource filter
IBaseFilter* CDXplay::FindSource(void)
{
	IEnumFilters*	EnumFilters;
	IBaseFilter*	FileSource;
	IEnumPins*		EnumPins;
	IPin*			InPin;
	int				ninputpins = 0;
	PIN_INFO		pinfo;

	if(!pGB)
		return NULL;

	OUTPUTDEBUG("DXplay: Search for source filter.");

	pGB->EnumFilters(&EnumFilters);
	EnumFilters->Reset();

	// go through all filters
	while(EnumFilters->Next(1, &FileSource, NULL) == S_OK)
	{
		FileSource->EnumPins(&EnumPins);
		EnumPins->Reset();
		ninputpins = 0;
		while(EnumPins->Next(1, &InPin, NULL) == S_OK) {
			InPin->QueryPinInfo(&pinfo);
			SAFE_RELEASE(pinfo.pFilter, "pinfo.pFilter")
			SAFE_RELEASE(InPin, "InPin")
			if(pinfo.dir == PINDIR_INPUT) {
				ninputpins++;
				break;		// we can jump out if we found an input pin
			}
		}
		SAFE_RELEASE(EnumPins, "EnumPins")
		if(ninputpins == 0) {
			SAFE_RELEASE(EnumFilters, "EnumFilters")
			return FileSource;
		} else
			SAFE_RELEASE(FileSource, "FileSource")
	}

	SAFE_RELEASE(EnumFilters, "EnumFilters")
	return NULL;
}


// Function removes splitter filter
IBaseFilter* CDXplay::FindSplitter(IBaseFilter *FileSource)
{
	IEnumPins*		EnumPins;
	IPin*			OutPin;
	IPin*			InPin;
	IBaseFilter		*DestFilter;
	PIN_INFO		pinfo;

	USES_CONVERSION;

	if(!FileSource)
		return FALSE;

	FileSource->EnumPins(&EnumPins);
	EnumPins->Reset();

	// FileSource only has 1 pin...
	if(EnumPins->Next(1, &OutPin, NULL) == S_OK) {

		OUTPUTDEBUG("DXplay: FileSource output pin found...");

		if(OutPin->ConnectedTo(&InPin) == S_OK) {
			SAFE_RELEASE(OutPin, "OutPin")
			InPin->QueryPinInfo(&pinfo);
			DestFilter = pinfo.pFilter;
			SAFE_RELEASE(InPin, "InPin")

#ifdef _DEBUG
			FILTER_INFO		pInfo;
			DestFilter->QueryFilterInfo(&pInfo);

			wsprintf(debug, TEXT("DXplay: MpegSplitter: %s found."), W2A(pInfo.achName));
			OutputDebugString(debug);
			SAFE_RELEASE(pInfo.pGraph, "pInfo.pGraph");
#endif
			SAFE_RELEASE(EnumPins, "EnumPins")
			return DestFilter;
		}
	}

	SAFE_RELEASE(EnumPins, "EnumPins")
	return FALSE;
}


void CDXplay::ReleaseAllFilters(void)
{
	IEnumFilters*	EnumFilters;
	IBaseFilter*	FileSource;

	OUTPUTDEBUG("DXplay: Search for filters...");

	pGB->EnumFilters(&EnumFilters);
	EnumFilters->Reset();

	while(EnumFilters->Next(1, &FileSource, NULL) == S_OK) {
		pGB->RemoveFilter(FileSource);
		EnumFilters->Reset();
		SAFE_RELEASE(FileSource, "FileSource");
	}

	SAFE_RELEASE(EnumFilters, "EnumFilters");

	OUTPUTDEBUG("DXplay: Filters removed...");
}


BOOL CDXplay::LoadMediaFile(const char* pszFileName)
{
	BOOL ret;

	IEnumFilters*	EnumFilters;
	IBaseFilter*	FileSource;
	IBaseFilter*	MpegSplitter;
	IBaseFilter*	pTempFilter;

	IEnumPins*		EnumPins;
	IPin*			InPin;

	USES_CONVERSION;

#ifdef _DEBUG
	FILTER_INFO		pInfo;

	wsprintf(debug, TEXT("DXplay: LoadMediaFile called: %s."), pszFileName);
	OutputDebugString(debug);
#endif

	if(!pGB) {
		OUTPUTDEBUG("DXplay: No graph built yet.");
		ret = RenderMediaFile(pszFileName);

#ifdef _DEBUG
		wsprintf(debug, TEXT("DXplay: RenderMediaFile returned: %d."), ret);
		OutputDebugString(debug);
		OutputDebugString("DXplay: Filters used: ");

		pGB->EnumFilters(&EnumFilters);
		EnumFilters->Reset();

		while(EnumFilters->Next(1, &FileSource, NULL) == S_OK) {
			FileSource->QueryFilterInfo(&pInfo);
			SAFE_RELEASE(FileSource, "FileSource")

			wsprintf(debug, TEXT("DXplay: %s"), W2A(pInfo.achName));
			OutputDebugString(debug);

			SAFE_RELEASE(pInfo.pGraph, "pInfo.pGraph");
		}

		SAFE_RELEASE(EnumFilters, "EnumFilters")

		OUTPUTDEBUG("DXplay: ----------------");
#endif

		return ret;
	}

	Stop();

	OUTPUTDEBUG("DXplay: Graph stopped, replacing source filter.");

	// For tier < T3, Stop() leaves filters on the graph
	// Try to save DShow
	// If there is an error removing filters...do render anyway
	while(tier < T3_REMOVEALL) {

		if(!(FileSource = FindSource())) {
			OUTPUTDEBUG("DXplay: FileSource filter NOT found.");
			break;
		}

		// Remove splitter as well
		if(tier > T0_REMOVESOURCE) {
			if(!(MpegSplitter = FindSplitter(FileSource))) {
				OUTPUTDEBUG("DXplay: Error removing splitter!");
				break;
			}

			// Disconnect all(!) pins
			if(tier == T2_DISCONNECT) {

				OUTPUTDEBUG("DXplay: Disconnecting all pins");

				pGB->EnumFilters(&EnumFilters);
				EnumFilters->Reset();

				// Find filters
				while(EnumFilters->Next(1, &pTempFilter, NULL) == S_OK) {
					pTempFilter->EnumPins(&EnumPins);
					EnumPins->Reset();

					// Find pins
					while(EnumPins->Next(1, &InPin, NULL) == S_OK) {
						InPin->Disconnect();
						SAFE_RELEASE(InPin, "InPin")
					}

					SAFE_RELEASE(EnumPins, "EnumPins")
					SAFE_RELEASE(pTempFilter, "pTempFilter")
				}

				SAFE_RELEASE(EnumFilters, "EnumFilters")
			}

			// Remove the splitter
			pGB->RemoveFilter(MpegSplitter);
			SAFE_RELEASE(MpegSplitter, "MpegSpliter");
			OUTPUTDEBUG("DXplay: MpegSplitter filter removed.");
		}

		// Finally, remove the source splitter
		pGB->RemoveFilter(FileSource);
		SAFE_RELEASE(FileSource, "FileSource");

		OUTPUTDEBUG("DXplay: FileSource filter removed.");
		break;
	}

	OUTPUTDEBUG("DXplay: Graph modifications complete. Rendering.");

	ret = RenderMediaFile(pszFileName);

#ifdef _DEBUG
	wsprintf(debug, TEXT("DXplay: RenderMediaFile returned: %d."), ret);
	OutputDebugString(debug);
#endif

	OUTPUTDEBUG("DXplay: LoadMediaFile completed.");

	return ret;
}


void CDXplay::ErrorReport(HRESULT hr)
{
	switch(hr) {
	case S_OK: 
		MessageBox(NULL, "Success.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_S_AUDIO_NOT_RENDERED: 
		MessageBox(NULL, "Partial success; the audio was not rendered. ", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_S_DUPLICATE_NAME:
		MessageBox(NULL, "Success; the Filter Graph Manager modified the filter name to avoid duplication.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_S_PARTIAL_RENDER:
		MessageBox(NULL, "Some of the streams in this movie are in an unsupported format.", NULL, MB_OK | MB_ICONERROR);
		break; 
	case VFW_S_VIDEO_NOT_RENDERED:
		MessageBox(NULL, "Partial success; some of the streams in this movie are in an unsupported format.", NULL, MB_OK | MB_ICONERROR);
		break;
	case E_ABORT:
		MessageBox(NULL, "Operation aborted.", NULL, MB_OK | MB_ICONERROR);
		break;
	case E_FAIL:
		MessageBox(NULL, "Failure.", NULL, MB_OK | MB_ICONERROR);
		break;
	case E_INVALIDARG:
		MessageBox(NULL, "Argument is invalid.", NULL, MB_OK | MB_ICONERROR);
		break;
	case E_OUTOFMEMORY:
		MessageBox(NULL, "Insufficient memory.", NULL, MB_OK | MB_ICONERROR);
		break;
	case E_POINTER:
		MessageBox(NULL, "NULL pointer argument.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_CANNOT_CONNECT:
		MessageBox(NULL, "No combination of intermediate filters could be found to make the connection.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_CANNOT_LOAD_SOURCE_FILTER:
		MessageBox(NULL, "The source filter for this file could not be loaded.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_CANNOT_RENDER:
		MessageBox(NULL, "No combination of filters could be found to render the stream.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_INVALID_FILE_FORMAT:
		MessageBox(NULL, "The file format is invalid.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_NOT_FOUND:
		MessageBox(NULL, "An object or name was not found.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_UNKNOWN_FILE_TYPE:
		MessageBox(NULL, "The media type of this file is not recognized.", NULL, MB_OK | MB_ICONERROR);
		break;
	case VFW_E_UNSUPPORTED_STREAM:
		MessageBox(NULL, "Cannot play back the file: the format is not supported.", NULL, MB_OK | MB_ICONERROR);
		break;
	default:
		char msg[256];
		wsprintf(msg, TEXT("Error Rendering file! (0x%x)"), hr);
		MessageBox(NULL, msg, NULL, MB_OK | MB_ICONERROR);
		break;
	}

}


BOOL CDXplay::GetVolume(long *db)
{
	OUTPUTDEBUG("DXplay: GetVolume called.");

	if(!db) {
		OUTPUTDEBUG("DXplay: GetVolume - Invalid parameters.");
		return FALSE;
	}

	if(!pBA) {
		OUTPUTDEBUG("DXplay: ERROR! No BasicAudio.");
		return FALSE;
	}

	pBA->get_Volume(db);

#ifdef _DEBUG
	wsprintf(debug, TEXT("DXplay: GetVolume returning %d dB."), db);
	OutputDebugString(debug);
#endif

	return TRUE;
}


BOOL CDXplay::SetVolume(long db)
{
	OUTPUTDEBUG("DXplay: SetVolume called.");

	if(!pBA)	{
		OUTPUTDEBUG("DXplay: ERROR! No BasicAudio.");
		return FALSE;
	}

#ifdef _DEBUG
	wsprintf(debug, TEXT("DXplay: Volume set at %d dB."), db);
	OutputDebugString(debug);
#endif

	return pBA->put_Volume(db);
}


BOOL CDXplay::IsRunning(void)
{
	OUTPUTDEBUG("DXplay: IsRunning called.");

	if(!pME) {
		OUTPUTDEBUG("DXplay: pME = NULL. Not running!");
		return FALSE;
	}

	if((hEvent == 0) && (pME->GetEventHandle((OAEVENT*) &hEvent) != S_OK)) {
		OUTPUTDEBUG("DXplay: Handle not retrieved. Not running.");
		return FALSE;
	}

	OUTPUTDEBUG("DXplay: Handle retrieved.");

	if(WaitForSingleObject(hEvent, 0) == WAIT_OBJECT_0) {
		long EventCode, Param1, Param2;

		OUTPUTDEBUG("DXplay: Event signaled.");

		while(pME->GetEvent(&EventCode, &Param1, &Param2, 0) != E_ABORT) {
			pME->FreeEventParams(EventCode, Param1, Param2);

			if(EC_COMPLETE == EventCode) {
				pME->FreeEventParams(EventCode, Param1, Param2);
				OUTPUTDEBUG("DXplay: Playback complete. Graph stopped.");
				return FALSE;			// Playback is complete
			}
		}
	}

	OUTPUTDEBUG("DXplay: Graph is running.");

	return TRUE;
}
